""" Python module to read Cameca (nano)SIMS data files. """

from sims.sims import *

__version__ = '1.0.0'
__name__ = 'sims'
__author__ = 'Zan Peeters'
__url__ = 'https://github.com/zanpeeters/sims'
__license__ = 'BSD 3-Clause Clear'
__copyright__ = '(c) 2018 Zan Peeters'
__description__ = __doc__
